import java.lang.reflect.Array
import groovy.json.JsonBuilder
import com.urbancode.anthill3.AHPTool
import com.urbancode.anthill3.plugin.jira.FailMode
import com.urbancode.anthill3.plugin.jira.JIRAHelper
import com.urbancode.anthill3.plugin.jira.XMLHelper
import org.apache.http.impl.client.CloseableHttpClient

final def ahptool = new AHPTool()
final def props = ahptool.getStepProperties()

final def serverUrl         = props['integration/serverUrl']
final def username          = props['integration/username']
final def password          = props['integration/passwordScript'] ?: props['integration/password']
final def transitionName    = props['transitionName']
final def additionalComment = props['additionalComment']
final def issueKeySingle    = props['issueKey']
final def issueKeyPattern   = props['issueKeyPattern']
final def resolutionName    = props['resolutionName']
final def failMode          = FailMode.valueOf(props['failMode'])

//------------------------------------------------------------------------------
// Script content
//------------------------------------------------------------------------------

println "Server: $serverUrl"
println "Transition: $transitionName"
println "Resolution: $resolutionName"
println "Issue Key: $issueKeySingle"
println "Issue Key Pattern: $issueKeyPattern"
println "Additional Comment: $additionalComment"
println "Fail mode configured for this step is ${failMode}"
println ""

def errorCount = 0
def issueIdSet = [] as Set

JIRAHelper jira = new JIRAHelper()
CloseableHttpClient client = jira.createClient(username, password)
try {
    XMLHelper helper = new XMLHelper(issueKeyPattern)
    helper.getChangesets()
    issueIdSet = helper.getIssueIdSet()

    issueIdSet.add(issueKeySingle)

    issueIdSet.each { issueKey ->
        // Confirm the updating issue exists
        if (!jira.doesIssueExist(client, serverUrl, issueKey)) {
            errorCount++
            def msg = "Skipped Issue ${issueKey}: Specified Issue not found."
            if (failMode == FailMode.FAIL_FAST) {
                throw new IllegalStateException("Fail Fast: " + msg)
            }
        }
        else {
            // Determine if specified transition is valid
            String transitionID = jira.getTransitionId(client, serverUrl, issueKey, transitionName)
            if (!transitionID) {
                errorCount++
                def msg = "Skipping Issue ${issueKey}: Specified Action \"${transitionName}\" was not available."
                println(msg);
                if (failMode == FailMode.FAIL_FAST) {
                    throw new IllegalStateException("Fail Fast: " + msg)
                }
            }
            else {
                // Issue exists and transition is valid
                // Create Update REST call body as HashMap object
                def updateIssueMap = [:]

                // Append the new transition state the issue will enter
                def transitionMap = [:]
                transitionMap.put("id", "${transitionID}")
                updateIssueMap.put("transition", transitionMap)

                // Confirm resolutionName is valid, if it has been set
                if (resolutionName && jira.isResolutionValid(client, serverUrl, issueKey, resolutionName)) {
                    def resolutionMap = [:]
                    resolutionMap.put("name", resolutionName)
                    def fieldsMap = [:]
                    fieldsMap.put("resolution", resolutionMap)
                    updateIssueMap.put("fields", fieldsMap)
                }

                // Append a comment that will be posted with the updated transition
                if (additionalComment) {
                    def addMap = [:]
                    addMap.put("body", additionalComment)
                    def commentMap = [:]
                    commentMap.put("add", addMap)

                    def updateMap = [:]
                    def commentArray = []
                    commentArray << commentMap
                    updateMap.put("comment", commentArray)
                    updateIssueMap.put("update", updateMap)
                }

                // Change mapping to JSON and prepare it to be sent via REST call
                JsonBuilder updateJSON = new JsonBuilder(updateIssueMap)

                // Call the do transitions REST call
                if (!jira.updateIssue(client, serverUrl, issueKey, updateJSON)) {
                    errorCount++
                    if(failMode == FailMode.FAIL_FAST) {
                        throw new IllegalStateException("Fail Fast: View above output for details.")
                    }
                }
                println "--------------------------"
            }
        }
    }
}
finally {
    client.close()
}

//------------------------------------------------------------------------------
// Check post conditions
//------------------------------------------------------------------------------

// Check if no issues were specified
if (failMode == FailMode.FAIL_ON_NO_UPDATES && !issueIdSet) {
    throw new IllegalStateException("Fail on no Updates: No issues specified.")
}

// Check if all updates failed
if (failMode == FailMode.FAIL_ON_NO_UPDATES && errorCount == issueIdSet.size()) {
    throw new IllegalStateException("Fail on no Updates: All issues failed to update.")
}

if (failMode == FailMode.FAIL_ON_ANY_FAILURE && errorCount > 0) {
    throw new IllegalStateException("Fail On Any Failure: Got ${errorCount} failures!")
}
